# Copyright 3D Control Systems, Inc. All Rights Reserved 2017-2019.
# Built in San Francisco.

# This software is distributed under a commercial license for personal,
# educational, corporate or any other use.
# The software as a whole or any parts of it is prohibited for distribution or
# use without obtaining a license from 3D Control Systems, Inc.

# All software licenses are subject to the 3DPrinterOS terms of use
# (available at https://www.3dprinteros.com/terms-and-conditions/),
# and privacy policy (available at https://www.3dprinteros.com/privacy-policy/)

import collections
import time
import threading

from base_sender import BaseSender

class Sender(BaseSender):

    MIN_AVERAGE_GCODE_TIME = 0.005 # 200 gcodes a second
    CALLBACK_DELAY = 2

    def __init__(self, parent, usb_info, profile):
        BaseSender.__init__(self, parent, usb_info, profile)
        self.print_thread = None
        self.printing_flag = False
        self.total_gcodes = 0
        self.lines = 0
        self.pause_flag = False
        self.cancel_flag = False
        self.buffer = collections.deque()

    def is_operational(self):
        #return False
        return not self.stop_flag

    def is_printing(self):
        return self.printing_flag

    def is_paused(self):
        return self.pause_flag

    def load_gcodes(self, gcodes):
        self.logger.info("Virtual printer loading gcodes")
        self.buffer = self.preprocess_gcodes(gcodes)
        self.lines = len(self.buffer)
        self.total_gcodes = self.lines
        self.print_thread = threading.Thread(target=self.printing, daemon=True)
        self.print_thread.start()

    def unbuffered_gcodes(self, gcodes):
        self.logger.info("Virtual printer unbuffered gcodes: %s" % gcodes)
        for gcode in self.preprocess_gcodes(gcodes):
            t = threading.Thread(target=self.delay_callback, args=("ok %s" % gcode, True), daemon=True) #switch here is you want success False
            t.start()

    def printing(self):
        self.printing_flag = True
        self.pause_flag = False
        self.current_line_number = 0
        # try to print any gcode in 100 seconds, but not faster then 200 gcodes per second
        delay_per_gcode = max(100.0 / self.total_gcodes, self.MIN_AVERAGE_GCODE_TIME)
        self.logger.info(f'Delay per gcode: {delay_per_gcode}')
        while not self.stop_flag and\
              not self.cancel_flag and\
              not self.current_line_number > self.total_gcodes: 
            if not self.pause_flag:
                try:
                    line = self.buffer[self.current_line_number]
                except IndexError:
                    break
                else:
                    line = line.split(b";")[0].strip() # remove any comments
                    if line:
                        response = "ok %s" % line
                        self.execute_callback(response, True)
                self.current_line_number += 1
            time.sleep(delay_per_gcode)
        self.printing_flag = False
        self.cancel_flag = False
        self.pause_flag = False
        self.buffer = self.buffer.clear()

    def cancel(self):
        self.logger.info("Virtual printer cancel")
        self.pause_flag = False
        self.cancel_flag = True
        time.sleep(self.MIN_AVERAGE_GCODE_TIME*2)

    def pause(self):
        self.logger.info("Virtual printer pause")
        self.pause_flag = True

    def unpause(self):
        self.logger.info("Virtual printer unpause")
        self.pause_flag = False

    def get_percent(self):
        if self.total_gcodes:
            return int(self.get_current_line_number() / float(self.total_gcodes) * 100)
        else:
            return 0

    def get_temps(self):
        percent = self.get_percent()
        return [percent + 10, percent + 20 * 2]

    def get_current_line_number(self):
        return self.current_line_number

    def reset(self):
        self.logger.info("Virtual printer reset")

    def delay_callback(self, line, success):
        time.sleep(self.CALLBACK_DELAY)
        super().execute_callback(line, success)

    def close(self):
        self.logger.info('Closing Virtual printer sender')
        self.stop_flag = True
        if self.print_thread:
            self.print_thread.join()
